/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package gr.abiss.mvn.plugins.jstools;

import gr.abiss.mvn.plugins.jstools.utils.FileSystemDirectoryUtils;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.ByteArrayOutputStream;
import java.io.PrintStream;
import java.util.Iterator;
import java.util.Locale;
import java.util.Set;
import org.apache.commons.io.IOUtils;
import org.apache.maven.reporting.MavenReportException;
import org.codehaus.doxia.sink.Sink;
import org.mozilla.javascript.Context;
import org.mozilla.javascript.Scriptable;

/**
 * Goal to generate JavaScript coverage reports using JSLint
 * 
 * @goal jslint
 * @phase post-site
 * @version $Id$
 * @author manos
 */
public class JsLintMojo extends AbstractBaseJstoolsReport {

	/**
	 * A path pointing to some custom JS code to drive JSLint via Rhino. You can provide your own,
	 * or let the plugin use its own internal version. 
	 * @parameter expression="${project.build.directory}/jslint/fulljslint_init.js"
	 */
	private String jslintInitJsFilePath;
	/**
	 * A path pointing to a full version of JSLint (e.g.
	 * /home/username/stuff/fulljslint.js). If no value is provided the plugin wil
	 * use its own internal version. You can find the file at 
	 * <a href="http://www.jslint.com/fulljslint.js">http://www.jslint.com/fulljslint.js</a>
	 * 
	 * @parameter expression="${project.build.directory}/jslint/fulljslint.js"
	 */
	private String jslintJsFilePath;
	
	/**
	 * Wether to only include errors in the report. Default is true (errors only)
	 * @parameter expression="true"
	 */
	private boolean reportJslintErrorsOnly;

	/**
	 * if use of some browser features should be restricted (unused JSLint
	 * option)
	 */
	boolean adsafe = true;

	/** if bitwise operators should not be allowed (JSLint option) */
	boolean bitwise = true;

	/** if the standard browser globals should be predefined (JSLint option) */
	boolean browser = true;

	/** if upper case HTML should be allowed (JSLint option) */
	boolean cap = true;

	/** if debugger statements should be allowed (JSLint option) */
	boolean debug = true;

	/** if === should be required (JSLint option) */
	boolean eqeqeq = true;

	/** if eval should be allowed (JSLint option) */
	boolean evil = true;

	/** if HTML fragments should be allowed (JSLint option) */
	boolean fragment = true;

	/** if line breaks should not be checked (JSLint option) */
	boolean laxbreak = true;

	/** if names should be checked (JSLint option) */
	boolean nomen = true;

	/** if the scan should stop on first error (JSLint option) */
	boolean passfail = true;

	/** if increment/decrement should not be allowed (JSLint option) */
	boolean plusplus = true;

	/** if the Rhino environment globals should be predefined (JSLint option) */
	boolean rhino = true;

	/** if variables should be declared before used (JSLint option) */
	boolean undef = true;

	/** if strict whitespace rules apply (JSLint option) */
	boolean white = true;

	/** if the Yahoo Widgets globals should be predefined (JSLint option) */
	boolean widget = true;

	/**
	 * Use JSLint to generate a code analysis report
	 * @see gr.abiss.mvn.plugins.jstools.AbstractBaseJstoolsReport#doGenerateReport(java.util.Locale)
	 */
	public void doGenerateReport(Locale defaultLocale) throws MavenReportException {
		try {
			File outDir = new File(this.getOutputDirectory());
			if (!outDir.exists()) {
				outDir.mkdirs();
			}
			Set jsFiles = this.getJavaScriptFiles();
			Sink sink = getSink();
			sink.head();
			sink.title();
			sink.text("JSLint Coverage Report");
			sink.title_();
			sink.head_();
			sink.body();
			sink.section1();
			sink.sectionTitle1();
			sink.text("Overview");
			sink.sectionTitle1_();
			sink.paragraph();
			sink.text("This report was build with JSLint, the JavaScript Verifier. " +
					"JSLint is a JavaScript program that looks for problems in other JavaScript programs.");
			sink.paragraph_();
			if (jsFiles.size() > 0) {
				String jsLintCode = getFileAsString(this.jslintJsFilePath) + this.getFileAsString(this.jslintInitJsFilePath);
				sink.paragraph();
				sink.text("Here is the list of files in this report:");
				sink.paragraph_();
				sink.list();
				for (Iterator iter = jsFiles.iterator(); iter.hasNext();) {
					File jsFile = (File) iter.next();
					sink.listItem();
					sink.rawText("<a href=\"#_"
							+ jsFile.getName().replaceAll(" ", "") + "\">"
							+ jsFile.getName() + "</a>");
					sink.listItem_();
				}
				sink.list_();
				sink.section1_();
				for (Iterator iter = jsFiles.iterator(); iter.hasNext();) {
					File jsFile = (File) iter.next();
					ByteArrayOutputStream buff = runJslintOnSource(jsLintCode, jsFile
							.getAbsolutePath());
					sink.rawText("<a id=\"_"
									+ jsFile.getName().replaceAll(" ", "")
									+ "\"> </a>");
					sink.sectionTitle1();
					sink.text(jsFile.getName());
					sink.sectionTitle1_();
					sink.rawText("<div class=\"source\"><pre>"
							+ (buff.size() > 0 ? buff.toString()
									: "No problems found") + "</pre></div>");
				}
			} else {
				sink.paragraph();
				sink.text("There are currently no files in this report.");
				sink.paragraph_();
			}
			sink.body_();
			sink.flush();
			sink.close();
		} catch (Exception e) {
			e.printStackTrace();
			throw new MavenReportException("Error building JSLint report", e);
		}
	}

	/**
	 * @param jsLintCode
	 * @param sourceFilePath
	 * @return
	 */
	private ByteArrayOutputStream runJslintOnSource(String jsLintCode, String sourceFilePath){
		ByteArrayOutputStream buff = new ByteArrayOutputStream();
		PrintStream out = new PrintStream(buff);
		Context context = Context.enter();
		Scriptable scope = context.initStandardObjects(null);
		Scriptable varOut = Context.toObject(out, scope);
		scope.put("out", scope, varOut);
		Scriptable varSourceFilePath = Context.toObject(sourceFilePath, scope);
		scope.put("sourceFilePath", scope, varSourceFilePath);
		Scriptable varReportJslintErrorsOnly = Context.toObject(this.reportJslintErrorsOnly, scope);
		scope.put("reportJslintErrorsOnly", scope, varReportJslintErrorsOnly);
		context.evaluateString(scope, jsLintCode,
				"mvn-jstools jslint script", 1, null);
		return buff;
	}

	/**
	 * Setup the plugin's internal JSLint version
	 * 
	 * @param defaultLocale
	 * @throws MavenReportException
	 */
	protected void setUp(Locale defaultLocale) throws MavenReportException {
		// String jrhinoSemiPath = "/jslint/rhino.js";
		// copy the internal jslint distribution archive to target/
		try {
			File destDir = new File(this.buildDir+"/jslint");
			destDir.mkdirs();
			copyScript("/jslint/fulljslint.js", this.buildDir+"/jslint/fulljslint.js");
			copyScript("/jslint/fulljslint_init.js", this.buildDir+"/jslint/fulljslint_init.js");
			// copyScript(jrhinoSemiPath, this.buildDir + jrhinoSemiPath );
		} catch (FileNotFoundException fe) {
			throw new MavenReportException("Cannot find resource in classpath", fe);
		} catch (IOException ioe) {
			throw new MavenReportException("Error copying resource to target dir", ioe);
		}
	}
	/**
	 * Tear down the plugin's internal JSLint version
	 * 
	 * @param defaultLocale
	 * @throws MavenReportException
	 */
	protected void tearDown(Locale defaultLocale) throws MavenReportException {
		try {
			File runDir = new File(this.buildDir+"/jslint");
			if(runDir.exists()){
				FileSystemDirectoryUtils.deleteDirectory(runDir);
			}
		} catch (Exception ioe) {
			throw new MavenReportException("Error clearing up", ioe);
		}
	}

	/**
	 * @param jslintSemiPath
	 * @throws FileNotFoundException
	 * @throws IOException
	 */
	private void copyScript(String classpath, String filePath)
			throws FileNotFoundException, IOException {
		InputStream input = this.getClass().getResourceAsStream(classpath);
		OutputStream output = new FileOutputStream(filePath);
		IOUtils.copy(input, output);
		input.close();
		output.close();
	}


	/**
	 * @see org.apache.maven.reporting.AbstractMavenReport#isExternalReport()
	 */
	public boolean isExternalReport() {
		return false;
	}

	

	/**
	 * @see gr.abiss.mvn.plugins.jstools.AbstractBaseJstoolsReport#getBundleKey()
	 */
	public String getBundleKey() {
		return "jslint";
	}
    
}
